"""Schema for neosctl."""

import configparser
import dataclasses
from typing import Optional, Union

from pydantic import BaseModel, Field, field_validator


class Auth(BaseModel):
    access_token: str = ""
    expires_in: Optional[int] = None
    refresh_token: str = ""
    refresh_expires_in: Optional[int] = None


class OptionalProfile(BaseModel):
    gateway_api_url: str = ""
    registry_api_url: str = ""
    iam_api_url: str = ""
    hub_api_url: str = ""
    storage_api_url: str = ""
    user: str = ""
    access_token: str = ""
    refresh_token: str = ""
    ignore_tls: bool = False
    account: str = ""


class Profile(OptionalProfile):
    gateway_api_url: str
    registry_api_url: str = ""
    iam_api_url: str = ""
    hub_api_url: str = ""
    storage_api_url: str
    user: str
    access_token: str
    refresh_token: str
    ignore_tls: bool
    account: str = "root"  # backwards compat default


class Core(BaseModel):
    name: str
    host: str
    account: Optional[str] = None
    active: bool = False

    @field_validator("account", mode="before")
    @classmethod
    def convert_null_account(cls: "type[Core]", v: str) -> Optional[str]:
        """Toml treats None as "null", convert back to None."""
        if v == "null":
            return None
        return v


class Env(BaseModel):
    name: str
    hub_api_url: str
    user: str
    access_token: str
    refresh_token: str
    ignore_tls: bool
    active: bool
    account: str = "root"
    cores: dict[str, Core] = Field(default_factory=dict)


class Credential(BaseModel):
    access_key_id: str
    secret_access_key: str


@dataclasses.dataclass
class Common:
    gateway_api_url: str
    hub_api_url: str
    storage_api_url: str
    profile_name: str
    config: configparser.ConfigParser
    env: dict
    credential: configparser.ConfigParser
    active_env: Optional[Env] = None
    active_core: Optional[Core] = None
    profile: Optional[Union[Profile, OptionalProfile]] = None

    @property
    def name(self) -> str:
        """Return name configuration."""
        if self.active_env:
            return self.active_env.name

        return self.profile_name

    @property
    def access_token(self) -> str:
        """Return access_token configuration."""
        if self.active_env:
            return self.active_env.access_token

        if self.profile:
            return self.profile.access_token

        return ""

    @property
    def refresh_token(self) -> str:
        """Return access_token configuration."""
        if self.active_env:
            return self.active_env.refresh_token

        if self.profile:
            return self.profile.refresh_token

        return ""

    @property
    def ignore_tls(self) -> bool:
        """Return ignore_tls configuration."""
        if self.active_env:
            return self.active_env.ignore_tls

        if self.profile:
            return self.profile.ignore_tls

        return False

    @property
    def account(self) -> str:
        """Return system account."""
        if self.active_env:
            return self.active_env.account

        if self.profile:
            return self.profile.account

        return "root"

    def get_gateway_api_url(self) -> str:
        """Return gateway api url.

        If a user profile is provided and defines a gateway url, return that,
        otherwise or fall back to cli defined default.
        """
        if self.active_core:
            return self.active_core.host

        if self.profile and self.profile.gateway_api_url:
            return self.profile.gateway_api_url
        return self.gateway_api_url

    def get_storage_api_url(self) -> str:
        """Return storage api url.

        If a user profile is provided and defines a storage url, return that,
        otherwise or fall back to cli defined default.
        """
        if self.active_core:
            return self.active_core.host.replace("://", "://saas.").replace("/api/gateway", "")

        if self.profile and self.profile.storage_api_url:
            return self.profile.storage_api_url
        return self.storage_api_url

    def get_hub_api_url(self) -> str:
        """Return hub api url.

        If a user profile is provided and defines a hub url, return that,
        otherwise fall back to cli defined default.
        """
        if self.active_env:
            return self.active_env.hub_api_url

        if self.profile and self.profile.hub_api_url:
            return self.profile.hub_api_url

        if self.profile and self.profile.iam_api_url:
            url = self.profile.iam_api_url
            url = url.replace("/api/iam", "/api/hub")
            return url.replace("/api/hub/iam", "/api/hub")

        return self.hub_api_url



"""Gateway common entities commands."""

from dataclasses import dataclass
from typing import Optional

import typer

from neosctl import constant, util
from neosctl.services.gateway import schema


@dataclass
class EntityArgGenerator:
    """Argument generator for entities."""

    entity_name: str

    @property
    def identifier(self):  # noqa: ANN201
        """Argument identifier."""
        return typer.Argument(..., help=f"{self.entity_name} identifier", callback=util.sanitize)

    @property
    def name(self):  # noqa: ANN201
        """Argument name."""
        return typer.Argument(..., help=f"{self.entity_name} name", callback=util.sanitize)

    @property
    def label(self):  # noqa: ANN201
        """Argument label."""
        return typer.Argument(..., help=f"{self.entity_name} label", callback=util.sanitize)

    @property
    def description(self):  # noqa: ANN201
        """Argument description."""
        return typer.Argument(..., help=f"{self.entity_name} description", callback=util.sanitize)

    @property
    def note(self):  # noqa: ANN201
        """Argument note."""
        return typer.Argument(..., help=f"{self.entity_name} note", callback=util.sanitize)

    @property
    def owner(self):  # noqa: ANN201
        """Argument owner."""
        return typer.Option(..., "--owner", "-o", help=f"{self.entity_name} owner", callback=util.sanitize)

    @property
    def owner_optional(self):  # noqa: ANN201
        """Argument optional owner."""
        return typer.Option(None, "--owner", "-o", help=f"{self.entity_name} owner", callback=util.sanitize)

    @property
    def contacts(self):  # noqa: ANN201
        """Argument contacts."""
        return typer.Option([], "--contact", "-c", help=f"{self.entity_name} contact IDs", callback=util.sanitize)

    @property
    def links(self):  # noqa: ANN201
        """Argument links."""
        return typer.Option([], "--link", "-l", help=f"{self.entity_name} links", callback=util.sanitize)


def create_entity(
    ctx: typer.Context,
    url: str,
    label: str,
    name: str,
    description: str,
    owner: Optional[str],
    contacts: list[str],
    links: list[str],
    entity_schema: Optional[schema.CreateEntity] = None,
) -> None:
    """Create entity."""
    if owner is None:
        if len(contacts) == 0 and len(links) == 0:
            info = None
        else:
            raise util.exit_with_output(
                msg="Set info fields: owner (required), contacts (optional) and links (optional).",
                exit_code=1,
            )
    else:
        info = schema.EntityInfo(
            owner=owner,
            contact_ids=contacts,
            links=links,
        )

    if entity_schema is None:
        entity_schema = schema.CreateEntity(
            name=name,
            label=label,
            description=description,
        )

    data = schema.CreateEntityRequest(
        entity=entity_schema,
        entity_info=info,
    )

    util.post_and_process(
        ctx,
        constant.GATEWAY,
        url,
        json=data.model_dump(exclude_none=True, by_alias=True),
    )


def list_entities(ctx: typer.Context, url: str) -> None:
    """List entities."""
    util.get_and_process(ctx, constant.GATEWAY, url)


def get_entity(ctx: typer.Context, url: str, identifier: str) -> None:
    """Get entity."""
    util.get_and_process(ctx, constant.GATEWAY, f"{url}/{identifier}")


def delete_entity(ctx: typer.Context, url: str, identifier: str) -> None:
    """Delete entity."""
    util.delete_and_process(ctx, constant.GATEWAY, f"{url}/{identifier}")


def update_entity(
    ctx: typer.Context,
    url: str,
    identifier: str,
    label: str,
    name: str,
    description: str,
    entity_schema: Optional[schema.CreateEntity] = None,
) -> None:
    """Update entity."""
    if entity_schema is None:
        entity_schema = schema.CreateEntity(
            name=name,
            label=label,
            description=description,
        )

    data = schema.UpdateEntityRequest(
        entity=entity_schema,
    )

    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{url}/{identifier}",
        json=data.model_dump(exclude_none=True, by_alias=True),
    )


def get_entity_info(ctx: typer.Context, url: str, identifier: str) -> None:
    """Get entity info."""
    util.get_and_process(ctx, constant.GATEWAY, f"{url}/{identifier}/info")


def update_entity_info(
    ctx: typer.Context,
    url: str,
    identifier: str,
    owner: str,
    contacts: list[str],
    links: list[str],
) -> None:
    """Update entity info."""
    data = schema.EntityInfo(
        owner=owner,
        contact_ids=contacts,
        links=links,
    )
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{url}/{identifier}/info",
        json=data.model_dump(exclude_none=True, by_alias=True),
    )


def get_entity_journal(
    ctx: typer.Context,
    url: str,
    identifier: str,
    page: int,
    per_page: int,
) -> None:
    """Get entity journal."""
    params = {"page": page, "per_page": per_page}
    util.get_and_process(ctx, constant.GATEWAY, f"{url}/{identifier}/journal", params=params)


def update_entity_journal(ctx: typer.Context, url: str, identifier: str, filepath: str) -> None:
    """Update entity journal note."""
    data = util.load_object(schema.UpdateJournalNote, filepath, "journal_note")

    util.post_and_process(
        ctx,
        constant.GATEWAY,
        f"{url}/{identifier}/journal",
        json=data.model_dump(by_alias=True),
    )


def get_entity_links(ctx: typer.Context, url: str, identifier: str) -> None:
    """Get entity links."""
    util.get_and_process(ctx, constant.GATEWAY, f"{url}/{identifier}/link")



"""Gateway journal note commands."""

import typer

from neosctl import constant, util
from neosctl.services.gateway import schema

app = typer.Typer(name="journal-note")


def _identified_journal_note_url(ctx: typer.Context, identifier: str) -> str:
    return "{}/v2/journal_note/{}".format(
        ctx.obj.get_gateway_api_url().rstrip("/"),
        identifier,
    )


IDENTIFIER_ARGUMENT = typer.Argument(..., help="Journal Note identifier", callback=util.sanitize)


@app.command(name="get")
def get_element(
    ctx: typer.Context,
    identifier: str = IDENTIFIER_ARGUMENT,
    _verbose: int = util.Verbosity,
) -> None:
    """Get journal note."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        _identified_journal_note_url(ctx, identifier),
    )


@app.command(name="update")
def update_element(
    ctx: typer.Context,
    identifier: str = IDENTIFIER_ARGUMENT,
    filepath: str = typer.Argument(..., help="Filepath to journal note payload", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update journal note."""
    data = util.load_object(schema.UpdateJournalNote, filepath, "journal_note")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        _identified_journal_note_url(ctx, identifier),
        json=data.model_dump(by_alias=True),
    )


@app.command(name="delete")
def delete_element(
    ctx: typer.Context,
    identifier: str = IDENTIFIER_ARGUMENT,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete journal note."""
    util.delete_and_process(
        ctx,
        constant.GATEWAY,
        _identified_journal_note_url(ctx, identifier),
    )



"""Schema for registry service."""

from pydantic import BaseModel


class RegisterCore(BaseModel):
    name: str
    public: bool


class MigrateCore(BaseModel):
    urn: str
    account: str


class AddCoreContact(BaseModel):
    user_id: str
    role: str


class RemoveCoreContact(BaseModel):
    user_id: str



"""Gateway data product commands."""

from typing import Optional

import typer

from neosctl import constant, util
from neosctl.services.gateway import entity, schema
from neosctl.util import prettify_json_no_sort

app = typer.Typer(name="data_product")


def _data_product_url(ctx: typer.Context) -> str:
    return "{}/v2/data_product".format(ctx.obj.get_gateway_api_url().rstrip("/"))


def _identified_data_product_url(ctx: typer.Context, identifier: str) -> str:
    return "{}/v2/data_product/{}".format(
        ctx.obj.get_gateway_api_url().rstrip("/"),
        identifier,
    )


arg_generator = entity.EntityArgGenerator("Data Product")


@app.command(name="create")
def create_entity(
    ctx: typer.Context,
    label: str = arg_generator.label,
    name: str = arg_generator.name,
    description: str = arg_generator.description,
    owner: Optional[str] = arg_generator.owner_optional,
    contacts: list[str] = arg_generator.contacts,
    links: list[str] = arg_generator.links,
    _verbose: int = util.Verbosity,
) -> None:
    """Create data product."""
    entity.create_entity(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        label=label,
        name=name,
        description=description,
        owner=owner,
        contacts=contacts,
        links=links,
    )


@app.command(name="list")
def list_entities(
    ctx: typer.Context,
    _verbose: int = util.Verbosity,
) -> None:
    """List data products."""
    entity.list_entities(ctx=ctx, url=_data_product_url(ctx=ctx))


@app.command(name="get")
def get_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product."""
    entity.get_entity(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="delete")
def delete_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete data product."""
    entity.delete_entity(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="get-info")
def get_entity_info(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product info."""
    entity.get_entity_info(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="update")
def update_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    label: str = arg_generator.label,
    name: str = arg_generator.name,
    description: str = arg_generator.description,
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product."""
    entity.update_entity(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
        label=label,
        name=name,
        description=description,
    )


@app.command(name="update-info")
def update_entity_info(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    owner: str = arg_generator.owner,
    contacts: list[str] = arg_generator.contacts,
    links: list[str] = arg_generator.links,
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product info."""
    entity.update_entity_info(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
        owner=owner,
        contacts=contacts,
        links=links,
    )


@app.command(name="get-journal")
def get_entity_journal(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    page: int = typer.Option(1, "--page", "-p", help="Page number", callback=util.sanitize),
    per_page: int = typer.Option(25, "--per-page", "-pp", help="Number of items per page", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product journal."""
    entity.get_entity_journal(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
        page=page,
        per_page=per_page,
    )


@app.command(name="update-journal")
def update_entity_journal(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to journal note payload", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product journal."""
    entity.update_entity_journal(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
        filepath=filepath,
    )


@app.command(name="get-links")
def get_entity_links(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product links."""
    entity.get_entity_links(
        ctx=ctx,
        url=_data_product_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="get-metadata")
def get_entity_metadata(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product metadata."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/metadata",
    )


@app.command(name="update-metadata")
def update_entity_metadata(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to metadata description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product metadata."""
    data = util.load_object(schema.UpdateEntityMetadataRequest, filepath, "metadata")

    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/metadata",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="delete-metadata")
def delete_entity_metadata(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to metadata description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Delete data product metadata."""
    data = util.load_object(schema.DeleteEntityMetadataRequest, filepath, "metadata")

    util.delete_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/metadata",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="update-schema")
def update_entity_schema(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to schema description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product schema."""
    data = util.load_object(schema.UpdateDataProductSchema, filepath, "schema")

    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/schema",
        json=data.model_dump(by_alias=True),
        timeout=20,
    )


@app.command(name="get-schema")
def get_entity_schema(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product schema."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/schema",
    )


@app.command(name="get-expectation-rules")
def get_entity_expectation_rules(
    ctx: typer.Context,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product expectation rules."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_data_product_url(ctx)}/expectation/rules",
    )


@app.command(name="get-expectation")
def get_entity_expectation(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    *,
    last_only: bool = typer.Option(
        False,
        "--last-only",
        "-l",
        help="Return only last settings.",
        callback=util.sanitize,
    ),
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product expectation settings."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/expectation?last_only={last_only}",
    )


@app.command(name="create-expectation-custom")
def create_expectation_custom(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to custom expectation description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Add data product custom expectation."""
    data = util.load_object(schema.ExpectationItem, filepath, "custom expectation")
    util.post_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/expectation/custom",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="update-expectation-custom")
def update_expectation_custom(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    custom_identifier: str = typer.Argument(..., help="Custom expectation identifier", callback=util.sanitize),
    filepath: str = typer.Argument(..., help="Filepath to custom expectation description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product custom expectation."""
    data = util.load_object(schema.ExpectationItem, filepath, "custom expectation")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/expectation/custom/{custom_identifier}",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="delete-expectation-custom")
def delete_expectation_custom(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    custom_identifier: str = typer.Argument(..., help="Custom expectation identifier", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Delete data product custom expectation."""
    util.delete_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/expectation/custom/{custom_identifier}",
    )


@app.command(name="update-expectation-weights")
def update_expectation_weights(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to expectation weights description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product expectation weighgts."""
    data = util.load_object(schema.ExpectationWeights, filepath, "expectation weights")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/expectation/weights",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="update-expectation-thresholds")
def update_expectation_thresholds(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to expectation thresholds description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product expectation thresholds."""
    data = util.load_object(schema.ExpectationThresholds, filepath, "expectation thresholds")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/expectation/thresholds",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="get-quality-profiling")
def get_entity_quality_profiling(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product profiling."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/profiling",
    )


@app.command(name="get-quality-validations")
def get_entity_quality_validations(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product validations."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/quality/validations",
    )


@app.command(name="get-classification-rule")
def get_data_product_classification_rule(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product classification rule."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/classification/rule",
    )


@app.command(name="update-classification-rule")
def update_data_product_classification_rule(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to classification rule description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product classification rule."""
    data = util.load_object(schema.ClassificationRule, filepath, "classification rule")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/classification/rule",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="get-classification-result")
def get_data_product_classification_result(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product classification result."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/classification/result",
    )


@app.command(name="update-classification-result")
def update_data_product_classification_result(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to classification rule description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product classification result."""
    data = util.load_object(schema.UpdateClassificationResult, filepath, "classification result")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/classification/result",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="get-lineage")
def get_entity_lineage(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product lineage."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/lineage",
    )


@app.command(name="get-data")
def get_entity_data(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product data."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/data",
    )


@app.command(name="delete-data")
def delete_entity_data(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete data product data."""
    util.delete_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/data",
    )


@app.command(name="publish")
def publish_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    *,
    private: bool = typer.Option(
        False,
        "--private",
        help="Limit visibility in mesh to core account.",
        callback=util.sanitize,
    ),
    _verbose: int = util.Verbosity,
) -> None:
    """Publish data product."""
    util.post_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/publish",
        params={"private": private},
    )


@app.command(name="unpublish")
def unpublish_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Unpublish data product."""
    util.delete_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/publish",
    )


@app.command(name="update-spark-file")
def update_entity_spark_file(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Spark job filepath", callback=util.sanitize),
    secrets: Optional[list[str]] = typer.Option(None, "--secret", "-s", help="Secret identifier"),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product spark file."""
    fp = util.get_file_location(filepath)

    with fp.open("rb") as f:
        util.put_and_process(
            ctx,
            constant.GATEWAY,
            f"{_identified_data_product_url(ctx, identifier)}/spark/file",
            files={"spark_file": f},
            params={"secret_identifiers": secrets} if secrets else None,
        )


@app.command(name="get-builder")
def get_entity_builder(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product builder."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/spark/builder",
        render_callable=prettify_json_no_sort,
    )


@app.command(name="get-spark-lineage")
def get_entity_spark_lineage(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product spark lineage."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/spark/lineage",
    )


@app.command(name="get-builder-state")
def get_entity_builder_state(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    _verbose: int = util.Verbosity,
) -> None:
    """Get data product builder state."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/spark/builder/state",
    )


@app.command(name="update-builder")
def update_entity_builder(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to builder description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product builder."""
    data = util.load_object(schema.BuilderPipeline, filepath, "builder")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/spark/builder",
        json=data.model_dump(by_alias=True),
        timeout=20,
    )


@app.command(name="update-spark-state")
def update_entity_spark_state(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to spark state description", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update data product spark state."""
    data = util.load_object(schema.UpdateSparkState, filepath, "state")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_product_url(ctx, identifier)}/spark/state",
        json=data.model_dump(by_alias=True),
    )



"""IAM service commands."""

import typing
from uuid import UUID

import httpx
import typer

from neosctl import constant, util
from neosctl.services.iam import schema
from neosctl.util import process_response, user_profile_callback

app = typer.Typer(name=constant.IAM, callback=user_profile_callback)

account_app = typer.Typer(name="account")
user_app = typer.Typer(name="user")
policy_app = typer.Typer(name="policy")
group_app = typer.Typer(name="group")

app.add_typer(account_app, name="account", help="Manage accounts.")
app.add_typer(user_app, name="user", help="Manage users.")
app.add_typer(group_app, name="group", help="Manage groups.")
app.add_typer(policy_app, name="policy", help="Manage policies.")


ACCOUNT_OPT = typer.Option(None, help="Account override (root only).", callback=util.sanitize)


def _iam_url(hub_api_url: str, postfix: str = "") -> str:
    return "{}/iam/{}".format(hub_api_url.rstrip("/"), postfix)


@account_app.command(name="create")
def create_account(
    ctx: typer.Context,
    display_name: str = typer.Option(..., "--display-name", "-d", help="Account display name."),
    name: str = typer.Option(..., "--name", "-n", help="Account name (used in urns)."),
    description: str = typer.Option(..., "--description", "--desc", help="Account description."),
    owner: str = typer.Option(..., "--owner", "-o", help="Account owner."),
    _verbose: int = util.Verbosity,
) -> None:
    """Create a system account."""

    @util.ensure_login
    def _request(ctx: typer.Context, data: schema.CreateAccount) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "account"),
            json=data.model_dump(mode="json"),
        )

    data = schema.CreateAccount(
        display_name=display_name,
        name=name,
        description=description,
        owner=owner,
    )

    r = _request(ctx, data)
    process_response(r)


@account_app.command(name="delete")
def delete_account(
    ctx: typer.Context,
    identifier: str = typer.Argument(..., help="Account identifier.", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Delete an account."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.delete(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"account/{identifier}"),
        )

    r = _request(ctx)
    process_response(r)


@account_app.command(name="update")
def update_account(
    ctx: typer.Context,
    display_name: str = typer.Option(..., "--display-name", "-d", help="Account display name."),
    description: str = typer.Option(..., "--description", "--desc", help="Account description."),
    owner: str = typer.Option(..., "--owner", "-o", help="Account owner."),
    identifier: str = typer.Argument(..., help="Account identifier.", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Update an account."""

    @util.ensure_login
    def _request(ctx: typer.Context, data: schema.UpdateAccount) -> httpx.Response:
        return util.put(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"account/{identifier}"),
            json=data.model_dump(mode="json"),
        )

    data = schema.UpdateAccount(
        display_name=display_name,
        description=description,
        owner=owner,
    )

    r = _request(ctx, data)
    process_response(r)


@account_app.command(name="list")
def list_accounts(
    ctx: typer.Context,
    _verbose: int = util.Verbosity,
) -> None:
    """List system accounts."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "account"),
        )

    r = _request(ctx)
    process_response(r)


@policy_app.command(name="list")
def list_policies(
    ctx: typer.Context,
    page: int = typer.Option(1, help="Page number."),
    page_size: int = typer.Option(10, help="Page size number."),
    resource: typing.Optional[str] = typer.Option(None, help="Resource nrn.", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """List existing policies."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        params: dict[str, typing.Union[int, str]] = {"page": page, "page_size": page_size}
        if resource:
            params["resource"] = resource

        return util.get(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "policy/users"),
            params=params,
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@policy_app.command(name="create")
def create_from_json(
    ctx: typer.Context,
    filepath: str = typer.Argument(..., help="Filepath of the user policy json payload", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Create an IAM policy."""

    @util.ensure_login
    def _request(ctx: typer.Context, user_policy: schema.UserPolicy) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            "{iam_url}".format(iam_url=_iam_url(ctx.obj.get_hub_api_url(), "policy/user")),
            json=user_policy.model_dump(mode="json"),
            account=account,
        )

    fp = util.get_file_location(filepath)
    user_policy_payload = util.load_json_file(fp, "policy")

    user_policy = schema.UserPolicy(**user_policy_payload)  # type: ignore[reportGeneralTypeIssues]

    r = _request(ctx, user_policy)
    process_response(r)


@policy_app.command(name="update")
def update_from_json(
    ctx: typer.Context,
    principal: str = typer.Argument(..., help="Principal uuid", callback=util.sanitize),
    filepath: str = typer.Argument(..., help="Filepath of the user policy json payload", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Update an existing IAM policy."""

    @util.ensure_login
    def _request(ctx: typer.Context, user_policy: schema.UserPolicy) -> httpx.Response:
        params = {"user_nrn": principal}
        return util.put(
            ctx,
            constant.IAM,
            "{iam_url}".format(iam_url=_iam_url(ctx.obj.get_hub_api_url(), "policy/user")),
            params=params,
            json=user_policy.model_dump(mode="json"),
            account=account,
        )

    fp = util.get_file_location(filepath)
    user_policy_payload = util.load_json_file(fp, "policy")

    user_policy = schema.UserPolicy(**user_policy_payload)  # type: ignore[reportGeneralTypeIssues]

    r = _request(ctx, user_policy)
    process_response(r)


@policy_app.command()
def delete(
    ctx: typer.Context,
    user_nrn: str = typer.Argument(..., callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete an existing IAM policy."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        params = {"user_nrn": user_nrn}
        return util.delete(
            ctx,
            constant.IAM,
            "{iam_url}".format(iam_url=_iam_url(ctx.obj.get_hub_api_url(), "policy/user")),
            params=params,
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@policy_app.command()
def get(
    ctx: typer.Context,
    user_nrn: str = typer.Argument(..., callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Get an existing IAM policy."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        params = {"user_nrn": user_nrn}
        return util.get(
            ctx,
            constant.IAM,
            "{iam_url}".format(iam_url=_iam_url(ctx.obj.get_hub_api_url(), "policy/user")),
            params=params,
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@user_app.command(name="create")
def create_user(
    ctx: typer.Context,
    username: str = typer.Option(..., "--username", "-u", callback=util.sanitize),
    email: str = typer.Option(..., "--email", "-e", callback=util.sanitize),
    first_name: str = typer.Option(..., "--first-name", "-f", callback=util.sanitize),
    last_name: str = typer.Option(..., "--last-name", "-l", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Create a keycloak user, and assign to account."""

    @util.ensure_login
    def _request(ctx: typer.Context, user: schema.CreateUser) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "user"),
            json=user.model_dump(mode="json"),
            account=account,
            timeout=30,
        )

    user = schema.CreateUser(
        enabled=True,
        username=username,
        email=email,
        first_name=first_name,
        last_name=last_name,
    )

    r = _request(ctx, user)
    process_response(r)


@user_app.command(name="delete")
def delete_user(
    ctx: typer.Context,
    user_id: str = typer.Option(
        ...,
        "--user-id",
        "-uid",
        help="User id in keycloak.",
        callback=util.sanitize,
    ),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Detach user from account."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.delete(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"user/{user_id}"),
            account=account,
            timeout=30,
        )

    r = _request(ctx)
    process_response(r)


@user_app.command(name="purge")
def purge_user(
    ctx: typer.Context,
    user_id: str = typer.Option(
        ...,
        "--user-id",
        "-uid",
        help="User id in keycloak.",
        callback=util.sanitize,
    ),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Purge user from core and IAM."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.delete(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"user/{user_id}/purge"),
            account=account,
            timeout=30,
        )

    r = _request(ctx)
    process_response(r)


@user_app.command(name="create-key-pair")
def create_key_pair(
    ctx: typer.Context,
    user_nrn: str = typer.Argument(..., callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Create an access key_pair and assign to a user."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"user/{user_nrn}/key_pair"),
            account=account,
            timeout=30,
        )

    r = _request(ctx)
    process_response(r)


@user_app.command(name="delete-key-pair")
def delete_key_pair(
    ctx: typer.Context,
    user_nrn: str = typer.Argument(..., callback=util.sanitize),
    access_key_id: str = typer.Argument(..., callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete the access key_pair from the user."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.delete(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"user/{user_nrn}/key_pair/{access_key_id}"),
            account=account,
            timeout=30,
        )

    r = _request(ctx)
    process_response(r)


@user_app.command(name="list")
def list_users(
    ctx: typer.Context,
    search: str = typer.Option(None, help="Search term", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """List existing keycloak users.

    Filter by search term on username, first_name, last_name, or email.
    """

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        params = {"search": search} if search else None
        return util.get(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "users"),
            params=params,
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@user_app.command(name="permissions")
def user_permissions(
    ctx: typer.Context,
    username: str = typer.Option(None, help="Keycloak username", callback=util.sanitize),
    identifier: UUID = typer.Option(None, help="User or Group identifier", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """List existing keycloak user permissions."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        user_id = identifier

        if username:
            params = {"search": username}
            r = util.get(
                ctx,
                constant.IAM,
                _iam_url(ctx.obj.get_hub_api_url(), "users"),
                params={"search": username},
                account=account,
            )
            if r.status_code >= constant.BAD_REQUEST_CODE:
                process_response(r)

            data = r.json()
            # In case search term matches email/name of another user, filter for specific username
            user_id = next((user["id"] for user in data["users"] if user["username"] == username), None)

        if user_id is None:
            typer.echo("User not found.")
            raise typer.Exit(code=1)

        params = {"user_nrn": user_id}
        return util.get(
            ctx,
            constant.IAM,
            "{iam_url}".format(iam_url=_iam_url(ctx.obj.get_hub_api_url(), "policy/user")),
            params=params,
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@user_app.command(name="reset-password")
def reset_password(
    ctx: typer.Context,
    username: str = typer.Argument(..., help="Keycloak user `username`", callback=util.sanitize),
    _verbose: int = util.Verbosity,
) -> None:
    """Request a password reset for a user."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "user/password/reset"),
            params={"username": username},
        )

    r = _request(ctx)
    process_response(r)


@group_app.command(name="create")
def create_group(
    ctx: typer.Context,
    name: str = typer.Option(..., help="Group name", callback=util.sanitize),
    description: str = typer.Option(..., help="Group description", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Create an IAM group."""

    @util.ensure_login
    def _request(ctx: typer.Context, group: schema.CreateUpdateGroup) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "group"),
            json=group.model_dump(mode="json"),
            account=account,
        )

    group = schema.CreateUpdateGroup(name=name, description=description)

    r = _request(ctx, group)
    process_response(r)


@group_app.command(name="update")
def update_group(
    ctx: typer.Context,
    identifier: UUID = typer.Argument(..., help="Group identifier", callback=util.sanitize),
    name: str = typer.Option(..., help="Group name", callback=util.sanitize),
    description: str = typer.Option(..., help="Group description", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Update an IAM group."""

    @util.ensure_login
    def _request(ctx: typer.Context, group: schema.CreateUpdateGroup) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"group/{identifier}"),
            json=group.model_dump(mode="json"),
            account=account,
        )

    group = schema.CreateUpdateGroup(name=name, description=description)

    r = _request(ctx, group)
    process_response(r)


@group_app.command(name="list")
def list_groups(
    ctx: typer.Context,
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """List IAM groups."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), "group"),
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@group_app.command(name="get")
def get_group(
    ctx: typer.Context,
    identifier: UUID = typer.Argument(..., help="Group identifier", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Get an IAM group."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"group/{identifier}"),
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@group_app.command(name="delete")
def delete_group(
    ctx: typer.Context,
    identifier: UUID = typer.Argument(..., help="Group identifier", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete an IAM group."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.delete(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"group/{identifier}"),
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@group_app.command(name="add-principals")
def add_principals(
    ctx: typer.Context,
    identifier: UUID = typer.Argument(..., help="Group identifier", callback=util.sanitize),
    principals: list[str] = typer.Option(
        ...,
        "--principal",
        "-p",
        help="Prinicpal identifiers",
        callback=util.sanitize,
    ),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Add principal(s) to an IAM group."""

    @util.ensure_login
    def _request(ctx: typer.Context, update: schema.Principals) -> httpx.Response:
        return util.post(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"group/{identifier}/principals"),
            json=update.model_dump(mode="json"),
            account=account,
        )

    update = schema.Principals(principals=principals)

    r = _request(ctx, update)
    process_response(r)


@group_app.command(name="remove-principals")
def remove_principals(
    ctx: typer.Context,
    identifier: UUID = typer.Argument(..., help="Group identifier", callback=util.sanitize),
    principals: list[str] = typer.Option(
        ...,
        "--principal",
        "-p",
        help="Prinicpal identifiers",
        callback=util.sanitize,
    ),
    account: typing.Optional[str] = ACCOUNT_OPT,
    _verbose: int = util.Verbosity,
) -> None:
    """Remove principal(s) from an IAM group."""

    @util.ensure_login
    def _request(ctx: typer.Context, update: schema.Principals) -> httpx.Response:
        return util.delete(
            ctx,
            constant.IAM,
            _iam_url(ctx.obj.get_hub_api_url(), f"group/{identifier}/principals"),
            json=update.model_dump(mode="json"),
            account=account,
        )

    update = schema.Principals(principals=principals)

    r = _request(ctx, update)
    process_response(r)



"""Gateway API schema definitions."""

from enum import Enum
from typing import Literal, Optional, Union

import pydantic
from pydantic import BaseModel


class EntityInfo(BaseModel):
    owner: str
    contact_ids: list[str]
    links: list[str]


class CreateElement(BaseModel):
    name: str


class CreateEntity(CreateElement):
    label: str
    description: str


class CreateOutput(CreateEntity):
    output_type: str


class CreateEntityRequest(BaseModel):
    entity: CreateEntity
    entity_info: Optional[EntityInfo]


class UpdateEntityRequest(BaseModel):
    entity: CreateEntity


class FieldMetadata(BaseModel):
    tags: Optional[list[str]]
    description: Optional[str]


class UpdateEntityMetadataRequest(BaseModel):
    fields: dict[str, FieldMetadata]
    tags: list[str]


class DeleteEntityMetadataRequest(BaseModel):
    fields: dict[str, FieldMetadata]
    tags: list[str]


class FieldDataType(pydantic.BaseModel):
    meta: dict[str, str]
    column_type: str


class CreateFieldDefinition(pydantic.BaseModel):
    name: str
    description: Optional[str] = None
    primary: bool = False
    optional: bool = False
    data_type: FieldDataType


class IcebergTableProperties(pydantic.BaseModel):
    table_format: str
    partitioning: Optional[list[str]] = None
    location: Optional[str] = None
    format_version: Optional[int] = None


class StreamingDataProductSchema(pydantic.BaseModel):
    product_type: Literal["streaming"]
    iceberg_table_properties: Optional[IcebergTableProperties] = None

    fields: list[CreateFieldDefinition]


class StoredDataProductSchema(pydantic.BaseModel):
    product_type: Literal["stored"]
    iceberg_table_properties: Optional[IcebergTableProperties] = None

    fields: list[CreateFieldDefinition]


class UpdateDataProductSchema(pydantic.BaseModel):
    details: Union[
        StreamingDataProductSchema,
        StoredDataProductSchema,
    ] = pydantic.Field(discriminator="product_type")


class ExpectationItem(BaseModel):
    expectation_type: str
    kwargs: dict
    meta: dict


class ExpectationColumnThresholds(BaseModel):
    accuracy: Optional[float]
    completeness: Optional[float]
    consistency: Optional[float]
    uniqueness: Optional[float]
    validity: Optional[float]


class ExpectationThresholds(BaseModel):
    table: float
    columns: dict[str, ExpectationColumnThresholds]


class ExpectationWeights(BaseModel):
    accuracy: float
    completeness: float
    consistency: float
    uniqueness: float
    validity: float


class UpdateQualityExpectations(BaseModel):
    custom_details: list[ExpectationItem]
    weights: Optional[ExpectationWeights]
    thresholds: Optional[ExpectationThresholds]


class ClassificationRegexRecognizer(BaseModel):
    name: str
    description: str
    label: str
    patterns: list[str]


class ClassificationRule(BaseModel):
    model: str
    excluded_columns: list[str]
    regex_recognizers: list[ClassificationRegexRecognizer]


class UpdateClassificationResult(BaseModel):
    resolve: list[str]


class BuilderPipeline(BaseModel):
    config: dict
    inputs: dict[str, dict]
    transformations: list
    finalisers: dict
    preview: bool = False


class UpdateSparkState(BaseModel):
    state: dict


class UpdateDataSourceConnection(BaseModel):
    connection: dict  # TODO: add more details when input parameters will be stable


class UpdateDataSourceConnectionSecret(BaseModel):
    secrets: dict  # TODO: this type is differ from the type in gateway


class UpdateDataUnitConfiguration(BaseModel):
    configuration: dict  # TODO: add more details when input parameters will be stable


class UpdateSecret(BaseModel):
    name: str
    data: dict


class UpdateJournalNote(BaseModel):
    note: str
    owner: str


class SecretKeys(BaseModel):
    keys: list[str]


class TagScope(Enum):
    schema = "SCHEMA"
    field = "FIELD"


class UpdateTag(BaseModel):
    tag: str
    scope: str



"""Profile commands, like profile initialisation, deletion, and listing."""

import re
import typing
from warnings import warn

import click
import typer

from neosctl import schema, util
from neosctl.util import (
    exit_with_output,
    get_user_profile_section,
    prettify_json,
    remove_config,
    upsert_config,
    upsert_credential,
)

app = typer.Typer()


r = re.compile(r"http[s]?:\/\/.*")


def _validate_url(value: str) -> str:
    """Validate a given url is a valid url with schema."""
    m = r.match(value)
    if m is None:
        msg = f"Invalid url, must match pattern: `{r}`."
        raise click.UsageError(msg)
    return value


@app.command()
def init(
    ctx: typer.Context,
    *,
    hostname: typing.Optional[str] = typer.Option(None, "--host", "-h", callback=util.sanitize),
    gateway_api_url: typing.Optional[str] = typer.Option(None, "--gateway-api-url", "-g", callback=util.sanitize),
    hub_api_url: typing.Optional[str] = typer.Option(None, "--hub-api-url", callback=util.sanitize),
    storage_api_url: typing.Optional[str] = typer.Option(None, "--storage-api-url", "-s", callback=util.sanitize),
    username: typing.Optional[str] = typer.Option(None, "--username", "-u", callback=util.sanitize),
    account: typing.Optional[str] = typer.Option("root", "--account", "-a", callback=util.sanitize),
    ignore_tls: bool = typer.Option(
        False,
        "--ignore-tls",
        help="Ignore TLS errors (useful in local/development environments",
    ),
    non_interactive: bool = typer.Option(
        False,
        "--non-interactive",
        help="Don't ask for input, generate api values based on hostname.",
    ),
    _verbose: int = util.Verbosity,
) -> None:
    """Initialise a profile.

    Create a profile that can be reused in later commands to define which
    services to interact with, and which user to interact as.

    Call `init` on an existing profile will update the existing profile.
    """
    warn(
        "neosctl profile is deprecated, please use neosctl env",
        FutureWarning,
        stacklevel=2,
    )
    typer.echo(f"Initialising [{ctx.obj.profile_name}] profile.")
    default_host = f"{hostname}/api/{{}}" if hostname else ""
    bare_host = hostname.replace("http://", "").replace("https://", "") if hostname else ""

    if non_interactive and not (hostname and username):
        raise exit_with_output(
            msg="\nError: --hostname/-h and --username/-u required for non-interactive mode.",
            exit_code=1,
        )

    if non_interactive:
        gateway_api_url = gateway_api_url or f"{hostname}/api/gateway"
        storage_api_url = storage_api_url or f"https://saas.{bare_host}"
        hub_api_url = hub_api_url or f"{hostname}/api/hub"

    urls = {
        "gateway_api_url": gateway_api_url,
        "hub_api_url": hub_api_url,
        "storage_api_url": storage_api_url,
    }
    for key, default, prompt in [
        ("gateway", ctx.obj.get_gateway_api_url(), "Gateway API url"),
        ("hub", ctx.obj.get_hub_api_url(), "Hub API url"),
        ("storage", ctx.obj.get_storage_api_url(), "Storage API url"),
    ]:
        url_key = f"{key}_api_url"
        url = urls[url_key]
        if url is None:
            urls[url_key] = typer.prompt(
                prompt,
                default=default or default_host.format(key),
                value_proc=_validate_url,
            )
        else:
            _validate_url(url)

    if username is None:
        kwargs = {}
        if ctx.obj.profile:
            kwargs["default"] = ctx.obj.profile.user
        username_input: str = typer.prompt(
            "Username",
            **kwargs,
        )
        username = username_input

    profile = schema.Profile(  # nosec: B106
        user=username,
        account=account,
        access_token="",
        refresh_token="",
        ignore_tls=ignore_tls,
        **urls,
    )

    upsert_config(ctx, profile)


@app.command()
def delete(
    ctx: typer.Context,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete a profile."""
    warn(
        "neosctl profile is deprecated, please use neosctl env",
        FutureWarning,
        stacklevel=2,
    )
    typer.confirm(f"Remove [{ctx.obj.profile_name}] profile", abort=True)
    remove_config(ctx)


@app.command()
def view(
    ctx: typer.Context,
    _verbose: int = util.Verbosity,
) -> None:
    """View configuration for a profile."""
    warn(
        "neosctl profile is deprecated, please use neosctl env",
        FutureWarning,
        stacklevel=2,
    )
    profile = get_user_profile_section(ctx.obj.config, ctx.obj.profile_name)
    raise exit_with_output(
        msg=prettify_json({k: v for k, v in profile.items()}),  # noqa: C416
    )


@app.command()
def credentials(
    ctx: typer.Context,
    access_key_id: str,
    secret_access_key: str,
    _verbose: int = util.Verbosity,
) -> None:
    """View configuration for a profile."""
    warn(
        "neosctl profile is deprecated, please use neosctl env",
        FutureWarning,
        stacklevel=2,
    )
    upsert_credential(ctx, access_key_id, secret_access_key)


@app.command(name="list")
def list_profiles(
    ctx: typer.Context,
    _verbose: int = util.Verbosity,
) -> None:
    """List available profiles."""
    warn(
        "neosctl profile is deprecated, please use neosctl env",
        FutureWarning,
        stacklevel=2,
    )
    raise exit_with_output(
        msg=prettify_json(sorted(ctx.obj.config.sections())),
    )



"""Storage API commands."""

import pathlib
import typing

import minio
import typer
from minio import S3Error
from minio.commonconfig import CopySource

from neosctl import util
from neosctl.util import exit_with_output, prettify_json

app = typer.Typer()

bucket_app = typer.Typer()
object_app = typer.Typer()
tagging_app = typer.Typer()

BUCKET_NAME_PATTERN = "^[a-z0-9][a-z0-9\\.\\-]{2,62}$"
OBJECT_NAME_PATTERN = "^[a-zA-Z0-9!\\ \\.\\-\\_*\\'\\(\\)]{1,255}$"

BUCKET_NAME_ARGUMENT = typer.Argument(
    ...,
    help="Bucket name",
    callback=util.validate_regex(BUCKET_NAME_PATTERN),
)

# Don't use regex as OBJECT_NAME can be `path/to/my/file.txt` etc. which could easily exceed 255 and break allowed chars
# minio just checks that object_name is not empty.
OBJECT_NAME_ARGUMENT = typer.Argument(
    ...,
    help="Object name",
    callback=util.validate_string_not_empty,
)

app.add_typer(bucket_app, name="bucket", help="Manage object buckets.")
app.add_typer(object_app, name="object", help="Manage objects.")
object_app.add_typer(tagging_app, name="tags", help="Manage object tags.")


def _generate_client(ctx: typer.Context) -> minio.Minio:
    credential = util.get_user_credential(ctx.obj.credential, ctx.obj.name)
    secure = ctx.obj.storage_api_url.startswith("https://")
    host = ctx.obj.storage_api_url.rstrip("/").replace("https://", "").replace("http://", "")

    return minio.Minio(  # nosec: B106
        host,
        access_key=credential.access_key_id,
        secret_key=credential.secret_access_key,
        secure=secure,
    )


def _s3error_to_json(e: S3Error) -> None:
    error = {
        "code": e.code,
        "message": e.message,
    }

    raise exit_with_output(
        msg=prettify_json(error),
        exit_code=1,
    )


@bucket_app.command(name="create")
def create_bucket(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    _verbose: int = util.Verbosity,
) -> None:
    """Create new bucket."""
    client = _generate_client(ctx)

    typer.echo(client.make_bucket(bucket_name))


@bucket_app.command(name="list")
def list_buckets(
    ctx: typer.Context,
    _verbose: int = util.Verbosity,
) -> None:
    """List buckets."""
    client = _generate_client(ctx)

    typer.echo([str(x) for x in client.list_buckets()])


@bucket_app.command(name="delete")
def delete_bucket(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete bucket."""
    client = _generate_client(ctx)
    typer.echo(client.remove_bucket(bucket_name))


@object_app.command(name="create")
def create_object(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    object_name: str = OBJECT_NAME_ARGUMENT,
    file: str = typer.Argument(
        ...,
        help="Path to the object file.",
        callback=util.validate_string_not_empty,
    ),
    _verbose: int = util.Verbosity,
) -> None:
    """Create object."""
    client = _generate_client(ctx)
    client.fput_object(
        bucket_name,
        object_name,
        file,
    )
    typer.echo(f"Object {object_name} is added to the bucket {bucket_name}")


@object_app.command(name="list")
def list_objects(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    prefix: typing.Union[str, None] = typer.Option(None, help="Path prefix"),
    *,
    recursive: bool = typer.Option(False, help="Recursively list bucket contents"),
    _verbose: int = util.Verbosity,
) -> None:
    """List objects."""
    client = _generate_client(ctx)
    typer.echo(
        [
            obj._object_name  # noqa: SLF001
            for obj in client.list_objects(bucket_name, prefix=prefix, recursive=recursive)
        ],
    )


@object_app.command(name="copy")
def copy_object(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    prefix: typing.Union[str, None] = typer.Option(None, help="Path prefix"),
    target_bucket_name: str = BUCKET_NAME_ARGUMENT,
    target_prefix: typing.Union[str, None] = typer.Option(None, help="Target path prefix"),
    _verbose: int = util.Verbosity,
) -> None:
    """List objects."""
    client = _generate_client(ctx)
    for obj in client.list_objects(bucket_name, prefix=prefix, recursive=True):
        if prefix is None and target_prefix is not None:
            # No prefix, prepend target prefix onto path `/path/to/file` -> `/target/path/to/file`
            target_object = f"{target_prefix.rstrip('/')}/{obj._object_name}"  # noqa: SLF001
        elif prefix is not None and target_prefix is not None:
            # Have prefix and target_prefix, replace prefix with target prefix `/prefix/file` -> `/target/file`
            target_object = obj._object_name.replace(prefix.rstrip("/"), target_prefix.rstrip("/"))  # noqa: SLF001
        else:
            # Have prefix and no target, or no prefix and no target `/path/to/file` -> `/path/to/file`
            target_object = obj._object_name  # noqa: SLF001

        client.copy_object(
            target_bucket_name,
            target_object,
            CopySource(
                bucket_name,
                obj._object_name,  # noqa: SLF001
            ),
        )


@object_app.command(name="get")
def get_object(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    object_name: str = OBJECT_NAME_ARGUMENT,
    file: str = typer.Argument(
        ...,
        help="Path to file where to store the object.",
        callback=util.validate_string_not_empty,
    ),
    _verbose: int = util.Verbosity,
) -> None:
    """Get object."""
    client = _generate_client(ctx)

    try:
        response = client.get_object(bucket_name, object_name)
    except S3Error as e:
        response = e.response
        _s3error_to_json(e)
        return
    else:
        data = response.data
    finally:
        response.close()
        response.release_conn()

    with pathlib.Path(file).open("wb") as fh:
        fh.write(data)


@object_app.command(name="delete")
def delete_object(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    object_name: str = OBJECT_NAME_ARGUMENT,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete object."""
    client = _generate_client(ctx)

    client.remove_object(bucket_name, object_name)
    typer.echo(f"Object {object_name} is deleted from the bucket {bucket_name}.")  # type: ignore[reportGeneralTypeIssues]


@tagging_app.command(name="set")
def set_object_tags(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    object_name: str = OBJECT_NAME_ARGUMENT,
    tags: list[str] = typer.Argument(
        ...,
        help="Tags as pairs of key=value",
        callback=util.validate_strings_are_not_empty,
    ),
    _verbose: int = util.Verbosity,
) -> None:
    """Set object tags. Be aware that this command overwrites any tags that are already set to the object."""
    client = _generate_client(ctx)

    minio_tags = minio.commonconfig.Tags.new_object_tags()  # type: ignore[reportGeneralTypeIssues]
    for tag in tags:
        key, value = tag.split("=", 1)
        minio_tags[key] = value
    client.set_object_tags(bucket_name, object_name, minio_tags)


@tagging_app.command(name="get")
def get_object_tags(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    object_name: str = OBJECT_NAME_ARGUMENT,
    _verbose: int = util.Verbosity,
) -> None:
    """Get object tags."""
    client = _generate_client(ctx)

    typer.echo(client.get_object_tags(bucket_name, object_name))


@tagging_app.command(name="delete")
def delete_object_tags(
    ctx: typer.Context,
    bucket_name: str = BUCKET_NAME_ARGUMENT,
    object_name: str = OBJECT_NAME_ARGUMENT,
    _verbose: int = util.Verbosity,
) -> None:
    """Delete object tags."""
    client = _generate_client(ctx)

    typer.echo(client.delete_object_tags(bucket_name, object_name))



"""Command line interface for neosctl."""

import configparser
import importlib.metadata
import logging
import os
import sys
import typing

import click
import typer
from rich.logging import RichHandler

from neosctl import (
    auth,
    constant,
    env,
    profile,
    schema,
    services,
    util,
)

logger_ = logging.getLogger(__name__)

VERBOSITY = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


def _generate_common_schema(
    profile_name: str,
    config: configparser.ConfigParser,
    credential: configparser.ConfigParser,
    env: configparser.ConfigParser,
    active_env: typing.Optional[schema.Env] = None,
    active_core: typing.Optional[schema.Core] = None,
    profile: typing.Union[schema.Profile, schema.OptionalProfile, None] = None,
) -> schema.Common:
    """Generate a common schema object.

    Set the default api values using user_profile and cli provided defaults.
    """
    common_schema = schema.Common(
        gateway_api_url="",
        hub_api_url="",
        storage_api_url="",
        profile_name=profile_name,
        config=config,
        credential=credential,
        env=env,
        profile=profile,
        active_env=active_env,
        active_core=active_core,
    )
    common_schema.gateway_api_url = common_schema.get_gateway_api_url()
    common_schema.hub_api_url = common_schema.get_hub_api_url()
    common_schema.storage_api_url = common_schema.get_storage_api_url()

    return common_schema


def _version_callback(*, value: bool) -> None:
    """Get current cli version."""
    if value:
        version = importlib.metadata.version("neosctl")
        typer.echo(f"neosctl {version}")
        raise typer.Exit


def setup_logging(verbose: int = 0) -> None:
    """Configure the logging."""
    logging.basicConfig(
        level=VERBOSITY.get(verbose, logging.DEBUG),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                rich_tracebacks=True,
                show_level=False,
                show_path=False,
                show_time=False,
                tracebacks_suppress=[click],
            ),
        ],
    )
    asyncio_logger = logging.getLogger("asyncio")
    asyncio_logger.disabled = True
    root_logger = logging.getLogger("")
    root_logger.setLevel(VERBOSITY.get(verbose, logging.DEBUG))


def _callback(
    ctx: typer.Context,
    version: typing.Optional[bool] = typer.Option(  # noqa: ARG001
        None,
        "--version",
        callback=_version_callback,
        help="Print version and exit.",
    ),
    profile: str = typer.Option(
        os.getenv("NEOSCTL_PROFILE", constant.DEFAULT_PROFILE),
        "--profile",
        "-p",
        help="Profile name",
        callback=util.sanitize,
    ),
    core: typing.Optional[str] = typer.Option(
        None,
        "--core",
        "-c",
        help="Active core override",
        callback=util.sanitize,
    ),
) -> None:
    """Inject common configuration defaults into context.

    Allow missing user_profile to support profile generation etc.
    """
    verbose = set(sys.argv) & {"-v", "-vv", "-vvv"}
    # If short verbosity is passed strip the `-` to get count of `v`s for verbosity setting.
    verbose = len(verbose.pop()) - 1 if verbose else 0

    # If full verbose flag is passed in, get count of flags
    if "--verbose" in sys.argv:
        verbose = sys.argv.count("--verbose")

    setup_logging(verbose)

    config = util.read_config_dotfile()
    env = util.read_env_dotfile()
    credential = util.read_credential_dotfile()
    user_profile = util.get_user_profile(config, profile, allow_missing=True)
    active_env = util.get_active_env(env)
    active_core = util.get_active_core(active_env, core)

    logger = logging.getLogger("neosctl.error")
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s | %(levelname)s | %(message)s")

    file_handler = logging.FileHandler(constant.LOG_FILEPATH)
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

    common_schema = _generate_common_schema(
        profile_name=profile,
        config=config,
        credential=credential,
        env=env,
        profile=user_profile,
        active_env=active_env,
        active_core=active_core,
    )

    ctx.obj = common_schema


app = typer.Typer(name="neosctl", callback=_callback)
app.add_typer(profile.app, name="profile", help="Manage profiles. [DEPRECATED]")
app.add_typer(
    auth.app,
    name="auth",
    callback=util.user_profile_callback,
    help="Manage authentication status. [DEPRECATED]",
)
app.add_typer(services.iam.app, name="iam", callback=util.user_profile_callback, help="Manage access policies.")
app.add_typer(
    services.storage.app,
    name="storage",
    callback=util.user_profile_callback,
    help="Interact with Storage (as a service).",
)
app.add_typer(
    services.gateway.app,
    name="gateway",
    callback=util.user_profile_callback,
    help="Interact with Gateway service.",
)
app.add_typer(
    services.registry.app,
    name="registry",
    callback=util.user_profile_callback,
    help="Manage cores and search data products.",
)

app.add_typer(env.app, name="env", help="Manage environments.")
